//---------------------------------------------------------------------------
//
//   CloneOf
//
//          Copyright(C) 2011, Claybird
//          Under New BSD License
//
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "MQPlugin.h"
#include <string>
#include <map>
#include <vector>
#include <set>


#ifndef COUNTOF
#define COUNTOF(x) (sizeof(x)/sizeof(x[0]))
#endif

HINSTANCE g_hInstance;
BOOL UpdateClones(MQDocument doc);

//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	g_hInstance=(HINSTANCE)hModule;
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    vOCIDԂB
//    ̊֐͋NɌĂяoB
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// v_Ng(Җ)IDAS64bit̒lƂĕԂ
	// l͑ƏdȂ悤ȃ_Ȃ̂ŗǂ
	*Product = 0xB3A8500C;
	*ID      = 0xc777497c;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    vOCԂB
//    ̊֐[vOCɂ]\ɌĂяoB
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// vOC
	return "CloneOf Ver.0.01      Copyright(C) 2011, Claybird";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    vOC̃^CvԂB
//    ̊֐͋NɌĂяoB
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// IuWFNgό`pvOCł
	return MQPLUGIN_TYPE_OBJECT;
}

//---------------------------------------------------------------------------
//  MQEnumString
//    |bvAbvj[ɕ\镶ԂB
//    ̊֐͋NɌĂяoB
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumString(int index)
{
	switch(index){
	case 0: return "Update Clones";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQModifyObject
//    j[IꂽƂɌĂяoB
//---------------------------------------------------------------------------

MQPLUGIN_EXPORT BOOL MQModifyObject(int index, MQDocument doc)
{
	switch(index){
	case 0: return UpdateClones(doc);
	}
	return FALSE;
}

//IuWFNg̈ʒů_擾
bool GetObjectBasePoint(MQObject obj,MQPoint& vBasePoint,int baseMaterialIdx)
{
	//---_}[J[p
	//ʎ擾
	std::vector<int> tmp;
	std::set<int> markerIndices;
	int numFaces=obj->GetFaceCount();
	for(int iFace=0;iFace<numFaces;iFace++){
		if(obj->GetFaceMaterial(iFace)==baseMaterialIdx){
			int numVert=obj->GetFacePointCount(iFace);
			if(numVert>0){
				tmp.resize(numVert);
				obj->GetFacePointArray(iFace,&tmp[0]);
				for(int j=0;j<numVert;j++){
					markerIndices.insert(tmp[j]);
				}
			}
		}
	}
	if(!markerIndices.empty()){
		//_擾
		std::vector<MQPoint> vertices;
		vertices.resize(obj->GetVertexCount());
		if(!vertices.empty()){
			obj->GetVertexArray(&vertices[0]);

			MQPoint vMarker;
			vMarker.zero();
			for(std::set<int>::iterator ite=markerIndices.begin();ite!=markerIndices.end();++ite){
				vMarker+=vertices[*ite];
			}
			vBasePoint=vMarker/(float)markerIndices.size();
		}
		return true;
	}
	return false;
}

//IuWFNg̒Sʒu擾
bool GetObjectCenter(MQObject obj,MQPoint& vBasePoint)
{
	//_擾
	std::vector<MQPoint> vertices;
	vertices.resize(obj->GetVertexCount());

	if(vertices.empty())return false;
	obj->GetVertexArray(&vertices[0]);

	//---IuWFNgAABBvZA̒SWԂ
	int activeVertices=0;

	MQPoint vMin,vMax;

	//---S_m
	for(size_t i=0;i<vertices.size();i++){
		if(obj->GetVertexRefCount(i)>0){
			if(activeVertices==0){
				vMin=vMax=vertices[i];
			}else{
				vMin.x=min(vMin.x,vertices[i].x);
				vMin.y=min(vMin.y,vertices[i].y);
				vMin.z=min(vMin.z,vertices[i].z);

				vMax.x=max(vMax.x,vertices[i].x);
				vMax.y=max(vMax.y,vertices[i].y);
				vMax.z=max(vMax.z,vertices[i].z);
			}
			activeVertices++;
		}
	}
	if(activeVertices==0)return false;

	vBasePoint=(vMin+vMax)*0.5f;

	return true;
}

//N[
void UpdateCloneWithOriginal(MQObject original,MQObject cloned,int nBaseMaterial)
{
	MQObject tmpObj=original->Clone();
	//t[Y
	tmpObj->Freeze(MQOBJECT_FREEZE_ALL);

	//sړ̐
	MQPoint vOrgCenter,vClonedCenter;

	if(
		//}[J[gʒuo
		(nBaseMaterial!=-1 && GetObjectBasePoint(tmpObj,vOrgCenter,nBaseMaterial) && GetObjectBasePoint(cloned,vClonedCenter,nBaseMaterial))
		//̂̒S
		|| (GetObjectCenter(tmpObj,vOrgCenter) && GetObjectCenter(cloned,vClonedCenter))
		){
		//---ό`s
		//TODO:gE]͍sȂ
		int count=tmpObj->GetVertexCount();
		for(int i=0;i<count;i++){
			if(tmpObj->GetVertexRefCount(i)>0){
				MQPoint vPos=tmpObj->GetVertex(i);
				vPos=vPos-vOrgCenter+vClonedCenter;
				tmpObj->SetVertex(i,vPos);
			}
		}
	}

	//N[NA
	{
		int count=cloned->GetVertexCount();
		for(int i=0;i<count;i++){
			cloned->DeleteVertex(i);
		}
	}

	//N[Ƀ}[W
	cloned->Merge(tmpObj);

	//N[̑|
	cloned->Compact();

	//n
	tmpObj->DeleteThis();
}

struct RELATION{
	int cloned;
	std::string strOriginalName;
};

//N[
BOOL UpdateClones(MQDocument doc)
{
	int numObjects=doc->GetObjectCount();
	if(numObjects <= 0)return FALSE;

	LPCTSTR lpTagName="cloneOf[";
	const char cCloseTag=']';
	const size_t tagLength=strlen(lpTagName);

	LPCTSTR lpBaseMaterialName="cloneOfMarker";

	//---IuWFNgꗗ:Oid̑Ή
	std::map<std::string,int> objectsMap;
	std::vector<RELATION> clones;

	for(int obj_index=0;obj_index<numObjects;obj_index++){
		MQObject obj = doc->GetObject(obj_index);
		if(obj==NULL)continue;
		//IuWFNĝ
		if(!obj->GetVisible())continue;
		char buffer[64];
		obj->GetName(buffer,COUNTOF(buffer));
		objectsMap[buffer]=obj_index;

		//N[Ώ(q)oꂽ
		if(_strnicmp(buffer,lpTagName,tagLength)==0){
			std::string strOriginal;

			const char* lpName=buffer+tagLength;
			for(;*lpName!=cCloseTag && *lpName!='\0';lpName++){
				strOriginal+=*lpName;
			}
			if(*lpName==cCloseTag && !strOriginal.empty()){
				//cloneOf[sourceObject]o:Ō']'ȂƂ͏ΏۊO
				RELATION rel;
				rel.cloned=obj_index;
				rel.strOriginalName=strOriginal;
				clones.push_back(rel);
			}
		}
	}

	//}[J[}eǍ
	int nBaseMaterial=-1;
	int numMaterials=doc->GetMaterialCount();
	for(int i=0;i<numMaterials;i++){
		MQMaterial mat=doc->GetMaterial(i);
		if(mat){
			char buffer[256];
			mat->GetName(buffer,COUNTOF(buffer));
			if(_stricmp(buffer,lpBaseMaterialName)==0){
				nBaseMaterial=i;
				break;
			}
		}
	}

	//---N[쐬
	BOOL bProcedded=FALSE;
	for(size_t i=0;i<clones.size();i++){
		std::map<std::string,int>::const_iterator ite=objectsMap.find(clones[i].strOriginalName);
		if(ite!=objectsMap.end()){
			int originalId=(*ite).second;
			MQObject original = doc->GetObject(originalId);
			MQObject cloned = doc->GetObject(clones[i].cloned);

			UpdateCloneWithOriginal(original,cloned,nBaseMaterial);
			bProcedded=TRUE;
		}
	}

	return bProcedded;
}


